/*
 *  cvLoadRGBJPEGBuffer.c
 *
 * Reads an RGB JPEG data buffer in memory and allocates/initialized a
 * OpenCV IplImage struct.
 *
 * To use: 
 * const unsigned char *jpeg_buf = ... // buffer containing jpeg data.
 * long jpeg_buf_size = ... // size of buffer
 * IplImage *img;
 * img = cvLoadRGBJPEGBuffer (jpeg_buf, jpeg_buf_size);
 * ... // Do stuff
 * cvReleaseImage (&img);
 *  
 * Copyright (C) 2009 Jeffrey Nelson
 * cvLoadRGBJPEGBuffer.c is based heavily on memsrc.c
 *
 *
 * memsrc.c
 *
 * Copyright (C) 1994-1996, Thomas G. Lane.
 * This file is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying README file.
 *
 * This file contains decompression data source routines for the case of
 * reading JPEG data from a memory buffer that is preloaded with the entire
 * JPEG file.  This would not seem especially useful at first sight, but
 * a number of people have asked for it.
 * This is really just a stripped-down version of jdatasrc.c.  Comparison
 * of this code with jdatasrc.c may be helpful in seeing how to make
 * custom source managers for other purposes.
 */
#include "cvLoadRGBJPEGBuffer.h"

#include <stdio.h>

/* this is not a core library module, so it doesn't define JPEG_INTERNALS */
#include <jpeglib.h>
#include <jerror.h>


/* Expanded data source object for memory input */

typedef struct
{
  struct jpeg_source_mgr pub;	/* public fields */

  JOCTET eoi_buffer[2];		/* a place to put a dummy EOI */
} my_source_mgr;

typedef my_source_mgr *my_src_ptr;


/*
 * Initialize source --- called by jpeg_read_header
 * before any data is actually read.
 */

void
init_source (j_decompress_ptr cinfo)
{
  /* No work, since jpeg_memory_src set up the buffer pointer and count.
   * Indeed, if we want to read multiple JPEG images from one buffer,
   * this *must* not do anything to the pointer.
   */
}


/*
 * Fill the input buffer --- called whenever buffer is emptied.
 *
 * In this application, this routine should never be called; if it is called,
 * the decompressor has overrun the end of the input buffer, implying we
 * supplied an incomplete or corrupt JPEG datastream.  A simple error exit
 * might be the most appropriate response.
 *
 * But what we choose to do in this code is to supply dummy EOI markers
 * in order to force the decompressor to finish processing and supply
 * some sort of output image, no matter how corrupted.
 */

static boolean
fill_input_buffer (j_decompress_ptr cinfo)
{
  my_src_ptr src = (my_src_ptr) cinfo->src;

  WARNMS (cinfo, JWRN_JPEG_EOF);

  /* Create a fake EOI marker */
  src->eoi_buffer[0] = (JOCTET) 0xFF;
  src->eoi_buffer[1] = (JOCTET) JPEG_EOI;
  src->pub.next_input_byte = src->eoi_buffer;
  src->pub.bytes_in_buffer = 2;

  return TRUE;
}


/*
 * Skip data --- used to skip over a potentially large amount of
 * uninteresting data (such as an APPn marker).
 *
 * If we overrun the end of the buffer, we let fill_input_buffer deal with
 * it.  An extremely large skip could cause some time-wasting here, but
 * it really isn't supposed to happen ... and the decompressor will never
 * skip more than 64K anyway.
 */

static void
skip_input_data (j_decompress_ptr cinfo, long num_bytes)
{
  my_src_ptr src = (my_src_ptr) cinfo->src;

  if (num_bytes > 0)
    {
      while (num_bytes > (long) src->pub.bytes_in_buffer)
	{
	  num_bytes -= (long) src->pub.bytes_in_buffer;
	  (void) fill_input_buffer (cinfo);
	  /* note we assume that fill_input_buffer will never return FALSE,
	   * so suspension need not be handled.
	   */
	}
      src->pub.next_input_byte += (size_t) num_bytes;
      src->pub.bytes_in_buffer -= (size_t) num_bytes;
    }
}


/*
 * An additional method that can be provided by data source modules is the
 * resync_to_restart method for error recovery in the presence of RST markers.
 * For the moment, this source module just uses the default resync method
 * provided by the JPEG library.  That method assumes that no backtracking
 * is possible.
 */


/*
 * Terminate source --- called by jpeg_finish_decompress
 * after all data has been read.  Often a no-op.
 *
 * NB: *not* called by jpeg_abort or jpeg_destroy; surrounding
 * application must deal with any cleanup that should happen even
 * for error exit.
 */

static void
term_source (j_decompress_ptr cinfo)
{
  /* no work necessary here */
}


/*
 * Prepare for input from a memory buffer.
 */

static void
jpeg_memory_src (j_decompress_ptr cinfo, const JOCTET * buffer,
		 size_t bufsize)
{
  my_src_ptr src;

  /* The source object is made permanent so that a series of JPEG images
   * can be read from a single buffer by calling jpeg_memory_src
   * only before the first one.
   * This makes it unsafe to use this manager and a different source
   * manager serially with the same JPEG object.  Caveat programmer.
   */
  if (cinfo->src == NULL)
    {				/* first time for this JPEG object? */
      cinfo->src = (struct jpeg_source_mgr *)
	(*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_PERMANENT,
				    sizeof (my_source_mgr));
    }

  src = (my_src_ptr) cinfo->src;
  src->pub.init_source = init_source;
  src->pub.fill_input_buffer = fill_input_buffer;
  src->pub.skip_input_data = skip_input_data;
  src->pub.resync_to_restart = jpeg_resync_to_restart;	/* use default method */
  src->pub.term_source = term_source;

  src->pub.next_input_byte = buffer;
  src->pub.bytes_in_buffer = bufsize;
}

/*
 * Returns an allocated/initialized IplImage from an in-memory jpeg buffer, 
 * or null on (some) error conditions.
 */
IplImage *
cvLoadRGBJPEGBuffer (const unsigned char *buf, long size)
{
  IplImage *imgBGR;
  struct jpeg_decompress_struct cinfo;
  struct jpeg_error_mgr jerr;
  unsigned char *raw_image;
  unsigned int i;
  unsigned long location = 0;
  /*
   * libjpeg data structure for storing one row (scanline) of an image
   */
  JSAMPROW row_pointer[1];

  if (buf == NULL || size < 1)
    {
      return NULL;
    }

  /*
   * set up the standard libjpeg error handler
   */
  cinfo.err = jpeg_std_error (&jerr);

  /*
   * initialize JPEG decompression object.
   */
  jpeg_create_decompress (&cinfo);

  /*
   * setup data source
   */
  jpeg_memory_src (&cinfo, buf, size);

  /*
   * reading the image header which contains image information
   */
  jpeg_read_header (&cinfo, TRUE);

  /*
   * Setup the IplImage header and allocate the bytes for the data.
   */
  imgBGR =
    cvCreateImageHeader (cvSize (cinfo.image_width, cinfo.image_height),
			 IPL_DEPTH_8U, cinfo.num_components);
  cvCreateData (imgBGR);

  /*
   * set raw image buffer pointer to the IplImage structure. 
   */
  raw_image = (unsigned char *) imgBGR->imageData;

  row_pointer[0] =
    (unsigned char *) malloc (cinfo.image_width * cinfo.num_components);


  /*
   * Begin decompressing image.
   */
  jpeg_start_decompress (&cinfo);

  /*
   * repeat for each scanline...
   */
  while (cinfo.output_scanline < cinfo.image_height)
    {
      jpeg_read_scanlines (&cinfo, row_pointer, 1);
      // Copy the raw RGB jpeg scanline data into the raw BGR IplImage data buf
      for (i = 0; i < cinfo.image_width * cinfo.num_components; i += 3)
	{
	  raw_image[location++] = row_pointer[0][i + 2];
	  raw_image[location++] = row_pointer[0][i + 1];
	  raw_image[location++] = row_pointer[0][i];
	}
    }

  /*
   * Complete decompressing operation since we are out of scanlines.
   */
  jpeg_finish_decompress (&cinfo);

  /*
   * Free up some memory...
   */
  jpeg_destroy_decompress (&cinfo);
  free (row_pointer[0]);
  return imgBGR;
}
